@@ -1,8 +1,6 @@
 # mypy: disable-error-code="empty-body"
 from __future__ import annotations
 
-import numpy as np
-
 from egglog import *
 
 from .array_api import *
@@ -13,27 +11,35 @@ from .program_gen import *
 # Depends on `np` as a global variable.
 ##
 
-array_api_module_string = Module([array_api_module.without_rules(), program_gen_module])
+array_api_program_gen_ruleset = ruleset(name="array_api_program_gen_ruleset")
+array_api_program_gen_eval_ruleset = ruleset(name="array_api_program_gen_eval_ruleset")
+
+array_api_program_gen_combined_ruleset = (
+    array_api_program_gen_ruleset
+    | program_gen_ruleset
+    | array_api_program_gen_eval_ruleset
+    | array_api_vec_to_cons_ruleset
+)
+array_api_program_gen_schedule = (array_api_program_gen_combined_ruleset | eval_program_rulseset).saturate()
 
 
-@array_api_module_string.function
-def bool_program(x: Boolean) -> Program:
-    ...
+@function
+def bool_program(x: Boolean) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _bool_program():
     yield rewrite(bool_program(TRUE)).to(Program("True"))
     yield rewrite(bool_program(FALSE)).to(Program("False"))
 
 
-@array_api_module_string.function()
-def int_program(x: Int) -> Program:
-    ...
+@function
+def int_program(x: Int) -> Program: ...
 
 
-@array_api_module_string.register
-def _int_program(i64_: i64, i: Int, j: Int):
+@array_api_program_gen_ruleset.register
+def _int_program(i64_: i64, i: Int, j: Int, s: String):
+    yield rewrite(int_program(Int.var(s))).to(Program(s, True))
     yield rewrite(int_program(Int(i64_))).to(Program(i64_.to_string()))
     yield rewrite(int_program(~i)).to(Program("~") + int_program(i))
     yield rewrite(bool_program(i < j)).to(Program("(") + int_program(i) + " < " + int_program(j) + ")")
@@ -53,55 +59,49 @@ def _int_program(i64_: i64, i: Int, j: Int):
     yield rewrite(int_program(i << j)).to(Program("(") + int_program(i) + " << " + int_program(j) + ")")
     yield rewrite(int_program(i >> j)).to(Program("(") + int_program(i) + " >> " + int_program(j) + ")")
     yield rewrite(int_program(i // j)).to(Program("(") + int_program(i) + " // " + int_program(j) + ")")
-    yield rewrite(int_program(Int(i64_))).to(Program(i64_.to_string()))
 
 
-@array_api_module_string.function()
-def tuple_int_program(x: TupleInt) -> Program:
-    ...
+@function
+def tuple_int_foldl_program(xs: TupleIntLike, f: Callable[[Program, Int], Program], init: ProgramLike) -> Program: ...
 
 
-@array_api_module_string.function()
-def tuple_int_program_inner(x: TupleInt) -> Program:
-    ...
+@function(ruleset=array_api_program_gen_ruleset)
+def tuple_int_program(x: TupleIntLike) -> Program:
+    return tuple_int_foldl_program(x, lambda acc, i: acc + int_program(i) + ", ", "(") + ")"
 
 
-@array_api_module_string.register
-def _tuple_int_program(i: Int, j: Int, ti: TupleInt, ti1: TupleInt, ti2: TupleInt):
+@array_api_program_gen_ruleset.register
+def _tuple_int_program(i: Int, ti: TupleInt, ti2: TupleInt, f: Callable[[Program, Int], Program], init: Program):
     yield rewrite(int_program(ti[i])).to(tuple_int_program(ti) + "[" + int_program(i) + "]")
+    yield rewrite(int_program(ti.length())).to(Program("len(") + tuple_int_program(ti) + ")")
 
-    yield rewrite(tuple_int_program(ti)).to(Program("(") + tuple_int_program_inner(ti) + ")")
-    yield rewrite(tuple_int_program_inner(ti1 + ti2)).to(
-        tuple_int_program_inner(ti1) + " " + tuple_int_program_inner(ti2)
-    )
-    yield rewrite(tuple_int_program_inner(TupleInt(i))).to(int_program(i) + ",")
+    yield rewrite(tuple_int_foldl_program(TupleInt.EMPTY, f, init)).to(init)
+    yield rewrite(tuple_int_foldl_program(ti.append(i), f, init)).to(f(tuple_int_foldl_program(ti, f, init), i))
 
-
-@array_api_module_string.function()
-def ndarray_program(x: NDArray) -> Program:
-    ...
-
-
-@array_api_module_string.function
-def ndarray_function_two(res: NDArray, l: NDArray, r: NDArray) -> Program:
-    ...
-
-
-@array_api_module_string.register
-def _ndarray_function_two(f: Program, res: NDArray, l: NDArray, r: NDArray, o: PyObject):
-    # When we have function, set the program and trigger it to be compiled
-    yield rule(eq(f).to(ndarray_function_two(res, l, r))).then(
-        union(f).with_(ndarray_program(res).function_two(ndarray_program(l), ndarray_program(r))),
-        f.eval_py_object({"np": np}),
+    yield rewrite(tuple_int_program(ti + ti2)).to(
+        Program("(") + tuple_int_program(ti) + " + " + tuple_int_program(ti2) + ")"
     )
 
 
-@array_api_module_string.function()
-def dtype_program(x: DType) -> Program:
-    ...
+@function
+def ndarray_program(x: NDArray) -> Program: ...
 
 
-@array_api_module_string.register
+@function(ruleset=array_api_program_gen_ruleset)
+def ndarray_function_two_program(res: NDArray, l: NDArray, r: NDArray) -> Program:
+    return ndarray_program(res).function_two(ndarray_program(l), ndarray_program(r))
+
+
+@function(ruleset=array_api_program_gen_eval_ruleset)
+def ndarray_function_two(res: NDArray, l: NDArray, r: NDArray) -> EvalProgram:
+    return EvalProgram(ndarray_function_two_program(res, l, r), {"np": np})
+
+
+@function
+def dtype_program(x: DType) -> Program: ...
+
+
+@array_api_program_gen_ruleset.register
 def _dtype_program():
     yield rewrite(dtype_program(DType.float64)).to(Program("np.dtype(np.float64)"))
     yield rewrite(dtype_program(DType.float32)).to(Program("np.dtype(np.float32)"))
@@ -111,13 +111,12 @@ def _dtype_program():
     yield rewrite(dtype_program(DType.object)).to(Program("np.dtype(np.object_)"))
 
 
-@array_api_module_string.function
-def float_program(x: Float) -> Program:
-    ...
+@function
+def float_program(x: Float) -> Program: ...
 
 
-@array_api_module_string.register
-def _float_program(f: Float, g: Float, f64_: f64, i: Int, r: Rational):
+@array_api_program_gen_ruleset.register
+def _float_program(f: Float, g: Float, f64_: f64, i: Int, r: BigRat):
     yield rewrite(float_program(Float(f64_))).to(Program(f64_.to_string()))
     yield rewrite(float_program(f.abs())).to(Program("np.abs(") + float_program(f) + ")")
     yield rewrite(float_program(Float.from_int(i))).to(int_program(i))
@@ -126,20 +125,20 @@ def _float_program(f: Float, g: Float, f64_: f64, i: Int, r: Rational):
     yield rewrite(float_program(f * g)).to(Program("(") + float_program(f) + " * " + float_program(g) + ")")
     yield rewrite(float_program(f / g)).to(Program("(") + float_program(f) + " / " + float_program(g) + ")")
     yield rewrite(float_program(Float.rational(r))).to(
-        Program("float(") + Program(r.numer.to_string()) + " / " + Program(r.denom.to_string()) + ")", r.denom != i64(1)
+        Program("float(") + Program(r.numer.to_string()) + " / " + Program(r.denom.to_string()) + ")",
+        ne(r.denom).to(BigInt(1)),
     )
     yield rewrite(float_program(Float.rational(r))).to(
-        Program("float(") + Program(r.numer.to_string()) + ")", eq(r.denom).to(i64(1))
+        Program("float(") + Program(r.numer.to_string()) + ")", eq(r.denom).to(BigInt(1))
     )
 
 
-@array_api_module_string.function()
-def value_program(x: Value) -> Program:
-    ...
+@function
+def value_program(x: Value) -> Program: ...
 
 
-@array_api_module_string.register
-def _value_program(i: Int, b: Boolean, f: Float, x: NDArray, v1: Value, v2: Value):
+@array_api_program_gen_ruleset.register
+def _value_program(i: Int, b: Boolean, f: Float, x: NDArray, v1: Value, v2: Value, xs: NDArray, ti: TupleInt):
     yield rewrite(value_program(Value.int(i))).to(int_program(i))
     yield rewrite(value_program(Value.bool(b))).to(bool_program(b))
     yield rewrite(value_program(Value.float(f))).to(float_program(f))
@@ -147,94 +146,96 @@ def _value_program(i: Int, b: Boolean, f: Float, x: NDArray, v1: Value, v2: Valu
     yield rewrite(value_program(x.to_value())).to(ndarray_program(x))
     yield rewrite(value_program(v1 < v2)).to(Program("(") + value_program(v1) + " < " + value_program(v2) + ")")
     yield rewrite(value_program(v1 / v2)).to(Program("(") + value_program(v1) + " / " + value_program(v2) + ")")
+    yield rewrite(value_program(v1 + v2)).to(Program("(") + value_program(v1) + " + " + value_program(v2) + ")")
+    yield rewrite(value_program(v1 * v2)).to(Program("(") + value_program(v1) + " * " + value_program(v2) + ")")
     yield rewrite(bool_program(v1.to_bool)).to(value_program(v1))
     yield rewrite(int_program(v1.to_int)).to(value_program(v1))
+    yield rewrite(value_program(xs.index(ti))).to((ndarray_program(xs) + "[" + tuple_int_program(ti) + "]").assign())
+    yield rewrite(value_program(v1.sqrt())).to(Program("np.sqrt(") + value_program(v1) + ")")
+    yield rewrite(value_program(v1.real())).to(Program("np.real(") + value_program(v1) + ")")
+    yield rewrite(value_program(v1.conj())).to(Program("np.conj(") + value_program(v1) + ")")
 
 
-@array_api_module_string.function()
-def tuple_value_program(x: TupleValue) -> Program:
-    ...
+@function
+def tuple_value_foldl_program(
+    xs: TupleValueLike, f: Callable[[Program, Value], Program], init: ProgramLike
+) -> Program: ...
 
 
-@array_api_module_string.function
-def tuple_value_program_inner(x: TupleValue) -> Program:
-    ...
+@function(ruleset=array_api_program_gen_ruleset)
+def tuple_value_program(x: TupleValueLike) -> Program:
+    return tuple_value_foldl_program(x, lambda acc, i: acc + value_program(i) + ", ", "(") + ")"
 
 
-@array_api_module_string.register
-def _tuple_value_program(tv1: TupleValue, tv2: TupleValue, v: Value):
-    yield rewrite(tuple_value_program(tv1)).to(Program("(") + tuple_value_program_inner(tv1) + ")")
-    yield rewrite(tuple_value_program_inner(tv1 + tv2)).to(
-        tuple_value_program_inner(tv1) + " " + tuple_value_program_inner(tv2)
-    )
-    yield rewrite(tuple_value_program_inner(TupleValue(v))).to(value_program(v) + ",")
+@array_api_program_gen_ruleset.register
+def _tuple_value_program(i: Int, ti: TupleValue, f: Callable[[Program, Value], Program], v: Value, init: Program):
+    yield rewrite(value_program(ti[i])).to(tuple_value_program(ti) + "[" + int_program(i) + "]")
+    yield rewrite(int_program(ti.length())).to(Program("len(") + tuple_value_program(ti) + ")")
+
+    yield rewrite(tuple_value_foldl_program(TupleValue.EMPTY, f, init)).to(init)
+    yield rewrite(tuple_value_foldl_program(ti.append(v), f, init)).to(f(tuple_value_foldl_program(ti, f, init), v))
 
 
-@array_api_module_string.function
-def tuple_ndarray_program(x: TupleNDArray) -> Program:
-    ...
+@function
+def tuple_ndarray_foldl_program(
+    xs: TupleNDArrayLike, f: Callable[[Program, NDArray], Program], init: ProgramLike
+) -> Program: ...
 
 
-@array_api_module_string.function
-def tuple_ndarray_program_inner(x: TupleNDArray) -> Program:
-    # Maps to terms seperated by commas, without other parens
-    ...
+@function(ruleset=array_api_program_gen_ruleset)
+def tuple_ndarray_program(x: TupleNDArrayLike) -> Program:
+    return tuple_ndarray_foldl_program(x, lambda acc, i: acc + ndarray_program(i) + ", ", "(") + ")"
 
 
-@array_api_module_string.register
-def _tuple_ndarray_program(x: NDArray, l: TupleNDArray, r: TupleNDArray, i: Int):
-    yield rewrite(tuple_ndarray_program(r)).to(Program("(") + tuple_ndarray_program_inner(r) + ")")
+@array_api_program_gen_ruleset.register
+def _tuple_ndarray_program(
+    i: Int, ti: TupleNDArray, f: Callable[[Program, NDArray], Program], v: NDArray, init: Program
+):
+    yield rewrite(ndarray_program(ti[i])).to(tuple_ndarray_program(ti) + "[" + int_program(i) + "]")
+    yield rewrite(int_program(ti.length())).to(Program("len(") + tuple_ndarray_program(ti) + ")")
 
-    yield rewrite(tuple_ndarray_program_inner(TupleNDArray(x))).to(ndarray_program(x) + ",")
-    yield rewrite(tuple_ndarray_program_inner(l + r)).to(
-        tuple_ndarray_program_inner(l) + " " + tuple_ndarray_program_inner(r)
-    )
-
-    yield rewrite(int_program(l.length())).to(Program("len(") + tuple_ndarray_program(l) + ")")
-    yield rewrite(ndarray_program(l[i])).to(tuple_ndarray_program(l) + "[" + int_program(i) + "]")
+    yield rewrite(tuple_ndarray_foldl_program(TupleNDArray.EMPTY, f, init)).to(init)
+    yield rewrite(tuple_ndarray_foldl_program(ti.append(v), f, init)).to(f(tuple_ndarray_foldl_program(ti, f, init), v))
 
 
-@array_api_module_string.function
-def optional_dtype_program(x: OptionalDType) -> Program:
-    ...
+@function
+def optional_dtype_program(x: OptionalDType) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _optional_dtype_program(dtype: DType):
     yield rewrite(optional_dtype_program(OptionalDType.none)).to(Program("None"))
     yield rewrite(optional_dtype_program(OptionalDType.some(dtype))).to(dtype_program(dtype))
 
 
-@array_api_module_string.function
-def optional_int_program(x: OptionalInt) -> Program:
-    ...
+@function
+def optional_int_program(x: OptionalInt) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _optional_int_program(x: Int):
     yield rewrite(optional_int_program(OptionalInt.none)).to(Program("None"))
     yield rewrite(optional_int_program(OptionalInt.some(x))).to(int_program(x))
 
 
-@array_api_module_string.function
+@function
 def optional_int_slice_program(x: OptionalInt) -> Program:
     """
     Translates an optional int to a program, but translates None as "" instead of None
     """
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _optional_int_slice_program(x: Int):
     yield rewrite(optional_int_slice_program(OptionalInt.none)).to(Program(""))
     yield rewrite(optional_int_slice_program(OptionalInt.some(x))).to(int_program(x))
 
 
-@array_api_module_string.function
-def slice_program(x: Slice) -> Program:
-    ...
+@function
+def slice_program(x: Slice) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _slice_program(start: OptionalInt, stop: OptionalInt, i: Int):
     yield rewrite(slice_program(Slice(start, stop, OptionalInt.none))).to(
         optional_int_slice_program(start) + ":" + optional_int_slice_program(stop)
@@ -244,12 +245,11 @@ def _slice_program(start: OptionalInt, stop: OptionalInt, i: Int):
     )
 
 
-@array_api_module_string.function
-def multi_axis_index_key_item_program(x: MultiAxisIndexKeyItem) -> Program:
-    ...
+@function
+def multi_axis_index_key_item_program(x: MultiAxisIndexKeyItem) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _multi_axis_index_key_item_program(i: Int, s: Slice):
     yield rewrite(multi_axis_index_key_item_program(MultiAxisIndexKeyItem.int(i))).to(int_program(i))
     yield rewrite(multi_axis_index_key_item_program(MultiAxisIndexKeyItem.slice(s))).to(slice_program(s))
@@ -257,57 +257,72 @@ def _multi_axis_index_key_item_program(i: Int, s: Slice):
     yield rewrite(multi_axis_index_key_item_program(MultiAxisIndexKeyItem.NONE)).to(Program("None"))
 
 
-@array_api_module_string.function
-def multi_axis_index_key_program(x: MultiAxisIndexKey) -> Program:
-    ...
+@function
+def multi_axis_index_key_program(x: MultiAxisIndexKey) -> Program: ...
 
 
-@array_api_module_string.register
-def _multi_axis_index_key_program(l: MultiAxisIndexKey, r: MultiAxisIndexKey, item: MultiAxisIndexKeyItem):
-    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey(item))).to(multi_axis_index_key_item_program(item))
-    yield rewrite(multi_axis_index_key_program(l + r)).to(
-        multi_axis_index_key_program(l) + ", " + multi_axis_index_key_program(r)
+@array_api_program_gen_ruleset.register
+def _multi_axis_index_key_program(
+    idx_fn: Callable[[Int], MultiAxisIndexKeyItem], k: i64, vec: Vec[MultiAxisIndexKeyItem], i: MultiAxisIndexKeyItem
+):
+    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey(0, idx_fn))).to(Program(""))
+
+    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey(Int(k), idx_fn))).to(
+        multi_axis_index_key_item_program(idx_fn(Int(0)))
+        + ", "
+        + multi_axis_index_key_program(MultiAxisIndexKey(Int(k - 1), lambda i: idx_fn(i + 1))),
+        ne(k).to(i64(0)),
+    )
+
+    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey.from_vec(Vec[MultiAxisIndexKeyItem]()))).to(
+        Program("")
+    )
+    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey.from_vec(vec))).to(
+        multi_axis_index_key_item_program(vec[0]) + ",",
+        eq(vec.length()).to(i64(1)),
+    )
+    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey.from_vec(vec))).to(
+        multi_axis_index_key_item_program(vec[0])
+        + ", "
+        + multi_axis_index_key_program(MultiAxisIndexKey.from_vec(vec.remove(0))),
+        vec.length() > 1,
     )
-    yield rewrite(multi_axis_index_key_program(MultiAxisIndexKey.EMPTY)).to(Program("()"))
 
 
-@array_api_module_string.function
-def index_key_program(x: IndexKey) -> Program:
-    ...
+@function
+def index_key_program(x: IndexKey) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _index_key_program(i: Int, s: Slice, key: MultiAxisIndexKey, a: NDArray):
     yield rewrite(index_key_program(IndexKey.ELLIPSIS)).to(Program("..."))
     yield rewrite(index_key_program(IndexKey.int(i))).to(int_program(i))
     yield rewrite(index_key_program(IndexKey.slice(s))).to(slice_program(s))
     yield rewrite(index_key_program(IndexKey.multi_axis(key))).to(multi_axis_index_key_program(key))
-    yield rewrite(index_key_program(ndarray_index(a))).to(ndarray_program(a))
+    yield rewrite(index_key_program(IndexKey.ndarray(a))).to(ndarray_program(a))
 
 
-@array_api_module_string.function
-def int_or_tuple_program(x: IntOrTuple) -> Program:
-    ...
+@function
+def int_or_tuple_program(x: IntOrTuple) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _int_or_tuple_program(x: Int, t: TupleInt):
     yield rewrite(int_or_tuple_program(IntOrTuple.int(x))).to(int_program(x))
     yield rewrite(int_or_tuple_program(IntOrTuple.tuple(t))).to(tuple_int_program(t))
 
 
-@array_api_module_string.function
-def optional_int_or_tuple_program(x: OptionalIntOrTuple) -> Program:
-    ...
+@function
+def optional_int_or_tuple_program(x: OptionalIntOrTuple) -> Program: ...
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _optional_int_or_tuple_program(it: IntOrTuple):
     yield rewrite(optional_int_or_tuple_program(OptionalIntOrTuple.some(it))).to(int_or_tuple_program(it))
     yield rewrite(optional_int_or_tuple_program(OptionalIntOrTuple.none)).to(Program("None"))
 
 
-@array_api_module_string.register
+@array_api_program_gen_ruleset.register
 def _ndarray_program(
     x: NDArray,
     y: NDArray,
@@ -323,6 +338,7 @@ def _ndarray_program(
     optional_device_: OptionalDevice,
     int_or_tuple_: IntOrTuple,
     idx: IndexKey,
+    odtype: OptionalDType,
 ):
     # Var
     yield rewrite(ndarray_program(NDArray.var(s))).to(Program(s, True))
@@ -478,8 +494,6 @@ def _ndarray_program(
     )
     # sqrt
     yield rewrite(ndarray_program(sqrt(x))).to((Program("np.sqrt(") + ndarray_program(x) + ")").assign())
-    # square
-    yield rewrite(ndarray_program(square(x))).to((Program("np.square(") + ndarray_program(x) + ")").assign())
     # Transpose
     yield rewrite(ndarray_program(x.T)).to(ndarray_program(x) + ".T")
     # sum
@@ -489,3 +503,8 @@ def _ndarray_program(
     )
     yield rewrite(tuple_int_program(x.shape)).to(ndarray_program(x) + ".shape")
     yield rewrite(ndarray_program(abs(x))).to((Program("np.abs(") + ndarray_program(x) + ")").assign())
+
+    # asarray
+    yield rewrite(ndarray_program(asarray(x, odtype, OptionalBool.none, optional_device_))).to(
+        Program("np.asarray(") + ndarray_program(x) + ", " + optional_dtype_program(odtype) + ")"
+    )
