@@ -1,13 +1,16 @@
-use std::collections::HashMap;
-
 use pyo3::prelude::*;
+use std::collections::HashMap;
+use std::collections::HashSet;
 
-#[pyclass(
-    unsendable,
-    text_signature = "(py_object_sort=None, *, fact_directory=None, seminaive=True, terms_encoding=False)"
-)]
+use crate::egraph::EGraph;
+
+#[pyclass()]
 pub struct SerializedEGraph {
     pub egraph: egraph_serialize::EGraph,
+    #[pyo3(get)]
+    pub truncated_functions: Vec<String>,
+    #[pyo3(get)]
+    pub discarded_functions: Vec<String>,
 }
 
 #[pymethods]
@@ -32,6 +35,13 @@ impl SerializedEGraph {
         serde_json::to_string(&self.egraph).unwrap()
     }
 
+    /// Split all primitive nodes, as well as other ops that match, into seperate e-classes
+    fn split_classes(&mut self, egraph: &EGraph, ops: HashSet<String>) {
+        self.egraph.split_classes(|id, node| {
+            egraph.egraph.from_node_id(id).is_primitive() || ops.contains(&node.op)
+        })
+    }
+
     /// Map each op name to a new op name.
     fn map_ops(&mut self, map: HashMap<String, String>) {
         for (_, node) in self.egraph.nodes.iter_mut() {
